/*
 *  fs.c
 *  NT Registry driver for Linux
 *
 *  Based on the ntfs driver by Martin Loewis, et.al.
 *
 *  Copyright (c) 1999, Todd Sabin

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/locks.h>
#include <linux/version.h>

#include <asm/uaccess.h>
#include <linux/types.h>
#include <linux/fs.h>

#include "ntreg.h"
#include "util.h"

#define min(a,b)  (((a)<(b))?(a):(b))

/*
 * Create a special file '.SD' in each directory containing the Security
 * Descriptor for that registy key?
 * FIXME: Should be made a mount option...
 */
#define DO_SK 1

/*
 * read in the Registry node for this inode.  The inode number is the
 * offset in the file (device) - 0x1000.  The first page is apparently
 * mostly junk.
 */
static void *
ntreg_read_node (kdev_t dev, unsigned long ino, int *err)
{
        struct buffer_head *bh;
        int block  = (ino+0x1000)/NTREG_BLOCK_SIZE;
        int offset = ino%NTREG_BLOCK_SIZE;
        int size;
        char *rc;
        char *dest;

	ntreg_debug(DEBUG_OTHER, "ntreg_read_node 0x%x\n", ino);

        if (!(bh = bread (dev, block, NTREG_BLOCK_SIZE))) {
                ntreg_error ("failed to read node %p\n", ino);
                *err = EIO;
                return NULL;
        }
        
        size = *(int *)((char*)bh->b_data+offset);
        if (size < 0) size = 0-size;
        if (!(rc = ntreg_malloc (size))) {
                ntreg_error ("failed to malloc for node %p\n", ino);
                brelse (bh);
                *err = ENOMEM;
                return NULL;
        }
        dest = rc;

        while  (offset+size > NTREG_BLOCK_SIZE) {
                memcpy (dest, (char*)bh->b_data+offset, NTREG_BLOCK_SIZE - offset);
                size -= NTREG_BLOCK_SIZE - offset;
                dest += NTREG_BLOCK_SIZE - offset;
                block++;
                offset = 0;
                brelse (bh);
                if (!(bh=bread (dev, block, NTREG_BLOCK_SIZE))) {
                        ntreg_error ("failed to read node %p\n", ino);
                        ntreg_free (rc);
                        *err = EIO;
                        return NULL;
                }
        }
        memcpy (dest, (char*)bh->b_data+offset, size);
        brelse (bh);
        *err = 0;
        return rc;
}


static ssize_t
ntreg_read (struct file * filp, char *buf, size_t count, loff_t *off)
{
        struct inode *ino = filp->f_dentry->d_inode;
        char *data = NULL;
        int len = 0;
        int err = 0;

        if (!ino || !(ino->i_mode&S_IFREG))
                return -EINVAL;

	ntreg_debug (DEBUG_OTHER, "ntreg_read 0x%x, 0x%x, 0x%x\n",
                     filp->f_dentry->d_inode->i_ino,
                     (unsigned)*off, (unsigned)count);

        if (VK_INO (ino)->magic == VK_MAGIC) {
                /* if data is elsewhere, load it */
                if (!VK_DATA_PRESENT (VK_INO (ino))) {
                        data = ntreg_read_node (ino->i_sb->s_dev,
                                                VK_INO (ino)->val_off,
                                                &err);
                        if (!data)
                                return -err;
                        data += 4;               /* skip over the block size */
                } else {
                        data = (char *)&VK_INO (ino)->val_off;
                }

                if (*off < ino->i_size) {
                        len = min (count, ino->i_size - *off);
                        copy_to_user (buf, data + *off, len);
                }

                /* free if needed */
                if (data != (char *)&VK_INO (ino)->val_off) {
                        ntreg_free (data-4);
                }
                *off+=len;
                return len;
#if DO_SK
        } else if (SK_INO (ino)->magic == SK_MAGIC) {
                if (*off < ino->i_size) {
                        len = min (count, ino->i_size - *off);
                        copy_to_user (buf, SK_INO (ino)->sk_data + *off, len);
                }
                *off+=len;
                return len;
#endif
        } else {
                ntreg_error ("ntreg_read: unknown data type for ino=0x%x\n",
                             ino->i_ino);
        }
}


static int
ntreg_add_dir (kdev_t dev, unsigned long ino,
               off_t pos, void *dirent, filldir_t filldir)
{
        int rc, err;
        char *slash;

        char *nk = ntreg_read_node (dev, ino, &err);
        if (!nk)
                return err;
        slash = memchr (NK (nk)->name, '/', NK (nk)->name_len);
        while (slash) {
                *slash = '\\';
                slash = memchr (NK (nk)->name, '/', NK (nk)->name_len);
        }
        rc = filldir (dirent, NK (nk)->name, NK (nk)->name_len, 
                      pos, ino);
        ntreg_free (nk);

        return rc;
}

static int
ntreg_add_file (kdev_t dev, unsigned long ino,
                off_t pos, void *dirent, filldir_t filldir)
{
        int rc, err;

        char *vk;
        char *slash;
        char *name;
        int name_len;

	ntreg_debug(DEBUG_OTHER, "ntreg_add_file: ino==%x\n",
                    ino);
        vk = ntreg_read_node (dev, ino, &err);
        if (!vk)
                return err;
        if (VK (vk)->name_len != 0) {
                slash = memchr (VK (vk)->name, '/', VK (vk)->name_len);
                while (slash) {
                        *slash = '\\';
                        slash = memchr (VK (vk)->name, '/', VK (vk)->name_len);
                }
                name = VK (vk)->name;
                name_len = VK (vk)->name_len;
        } else {
                name = "Unnamed-Value";
                name_len = 13;
        }
        rc = filldir (dirent, name, name_len, pos, ino);
        ntreg_free (vk);

        return rc;
}


static int
ntreg_readdir_lf (struct file *filp, void *data, 
                  struct inode *inode, int base,
                  void *dirent, filldir_t filldir)
{
        int i, rc = 0;
        for (i = filp->f_pos-base; i < LF (data)->num_child; i++) {
                if (rc = ntreg_add_dir (inode->i_sb->s_dev, 
                                        LF (data)->u[i].child_off, 
                                        filp->f_pos, dirent, filldir)) {
                        break;
                }
                filp->f_pos++;
        }
        return rc;
}

static int
ntreg_readdir_li (struct file *filp, void *data, 
                  struct inode *inode, int base,
                  void *dirent, filldir_t filldir)
{
        int i, rc = 0;
        for (i = filp->f_pos-base; i < LI (data)->num_child; i++) {
                if (rc = ntreg_add_dir (inode->i_sb->s_dev, 
                                        LI (data)->u[i].child_off, 
                                        filp->f_pos, dirent, filldir)) {
                        break;
                }
                filp->f_pos++;
        }
        return rc;
}

static int
ntreg_readdir_ri (struct file *filp, void *data, 
                  struct inode *inode, int base,
                  void *dirent, filldir_t filldir)
{
        int i, rc = 0, err;
        char *sub_data = NULL;

        for (i=0; i < RI (data)->num_child; i++) {
                sub_data = ntreg_read_node (inode->i_sb->s_dev, 
                                            RI (data)->u[i].child_off,
                                            &err);
                if (!sub_data)
                        return err;
                if (IS_LI (sub_data)) {
                        if (rc = ntreg_readdir_li (filp, sub_data,
                                                   inode, base, dirent,
                                                   filldir)) {
                                break;
                        }
                        base += LI (sub_data)->num_child;
                } else {
                        ntreg_error ("Invalid RI entry for ino=0x%x",
                                     inode->i_ino);
                        rc = EINVAL; /* FIXME:  better error? */
                }
                ntreg_free (sub_data);
                sub_data = NULL;
        }
        if (sub_data)
                ntreg_free (sub_data);

        return rc;
}

/*
 * Need to know how many 'special' directory entries we add automatically.
 */
#if DO_SK
#define NUM_SPECIAL 3
#else
#define NUM_SPECIAL 2
#endif
/* readdir returns '.', then '..', then the directory entries in sequence
   As the root directory contains a entry for itself, '.' is not emulated
   for the root directory */
static int
ntreg_readdir (struct file* filp, void *dirent, filldir_t filldir)
{
        int i, err;
	struct inode *inode = filp->f_dentry->d_inode;

	ntreg_debug(DEBUG_OTHER, "ntreg_readdir: filp==%p, dirent==%p, filldir==%p, inode==%p\n",
                    filp, dirent, filldir, inode->i_ino);

	if (!inode || !S_ISDIR(inode->i_mode))
		return -EBADF;

	if (filp->f_pos < NUM_SPECIAL) {
		if (filp->f_pos < 1) {
			if (filldir (dirent, ".", 1, 0, inode->i_ino) < 0)
				return 0;
			filp->f_pos ++;
		}
		if (filp->f_pos < 2) {
                        if (filldir (dirent, "..", 2, 1,
                                     (inode->i_ino == 0x20) ?
                                     inode->i_ino : NK_INO (inode)->parent_off) < 0)
                                return 0;
                        filp->f_pos ++;
                }
#if DO_SK
		if (filldir (dirent, ".SD", 3, 2, NK_INO (inode)->sk_off) < 0)
			return 0;
		filp->f_pos++;
#endif
	}
        
        /*
         * do directories
         */
        err = 0;
        if ((NK_INO(inode)->num_lf > 0)
            && (filp->f_pos-NUM_SPECIAL < NK_INO(inode)->num_lf)) {
                /* read in lf node */
                char *data = ntreg_read_node (inode->i_sb->s_dev, 
                                              NK_INO(inode)->lf_off, &err);
                ntreg_debug(DEBUG_OTHER, "ntreg_readdir: doing dirs\n");

                if (!data)
                        return -err;
                if (IS_LF (data)) {
                        err = ntreg_readdir_lf (filp, data, inode, 
                                                NUM_SPECIAL, dirent, filldir);
                } else if (IS_LI (data)) {
                        err = ntreg_readdir_li (filp, data, inode, 
                                                NUM_SPECIAL, dirent, filldir);
                } else if (IS_RI (data)) {
                        err = ntreg_readdir_ri (filp, data, inode, 
                                                NUM_SPECIAL, dirent, filldir);
                } else {
                        ntreg_error ("Invalid leaf entry for ino=0x%x",
                                     inode->i_ino);
                }
                ntreg_free (data);
        }
        if (err < 0)
                return 0;
        if (err)
                return -err;

        /*
         * then files
         */
        if ((NK_INO (inode)->num_val > 0)
            && (filp->f_pos - NUM_SPECIAL >= NK_INO (inode)->num_lf)
            && (filp->f_pos - NUM_SPECIAL < NK_INO (inode)->num_lf + NK_INO (inode)->num_val)) {
                /* read in val_arr */
                long *val_arr = ntreg_read_node (inode->i_sb->s_dev, 
                                                 NK_INO (inode)->val_arr_off,
                                                 &err);

                ntreg_debug(DEBUG_OTHER, "ntreg_readdir: doing files\n");

                if (!val_arr)
                        return -err;
                for (i = filp->f_pos- NUM_SPECIAL - NK_INO (inode)->num_lf;
                     i < NK_INO (inode)->num_val; 
                     i++) {
                        ntreg_debug(DEBUG_OTHER, "ntreg_readdir: doing file[%d]\n", i);
                        if (err = ntreg_add_file (inode->i_sb->s_dev, 
                                                  val_arr[i+1],  /* +1 to skip over size entry */
                                                  filp->f_pos, dirent, 
                                                  filldir))
                                break;
                        filp->f_pos++;
                }
                ntreg_free (val_arr);
        }
        if (err <= 0)
                return 0;
        if (err)
                return -err;
}

			
static int
ntreg_cmp_dir (kdev_t dev, unsigned long ino, const char *name)
{
        int rc=-1, err;
        int len;
        char *slash;

        char *nk = ntreg_read_node (dev, ino, &err);
        if (!nk)
                return err;
        len = strlen (name);
        if (len == NK (nk)->name_len) {
                slash = memchr (NK (nk)->name, '/', NK (nk)->name_len);
                while (slash) {
                        *slash = '\\';
                        slash = memchr (NK (nk)->name, '/', NK (nk)->name_len);
                }
                if (strncmp (name, NK (nk)->name, NK (nk)->name_len) == 0)
                        rc = 0;
        } 
        ntreg_free (nk);

        return rc;
}

static int
ntreg_cmp_file (kdev_t dev, unsigned long ino, const char *fname)
{
        int rc, err;
        int len;
        char *slash;
        char *name;
        int name_len;

        char *vk = ntreg_read_node (dev, ino, &err);
        if (!vk)
                return err;
        len = strlen (fname);
        if (VK (vk)->name_len != 0) {
                slash = memchr (VK (vk)->name, '/', VK (vk)->name_len);
                while (slash) {
                        *slash = '\\';
                        slash = memchr (VK (vk)->name, '/', VK (vk)->name_len);
                }
                name = VK (vk)->name;
                name_len = VK (vk)->name_len;
        } else {
                name = "Unnamed-Value";
                name_len = 13;
        }
        if (len == name_len &&
            strncmp (fname, name, name_len) == 0)
                rc = 0;
        else
                rc = -1;

        ntreg_free (vk);

        return rc;
}


static long
ntreg_lookup_lf (struct inode *dir, struct dentry *d, char *data)
{
        int i, rc;
        for (i = 0; i < LF (data)->num_child; i++) {
                rc = ntreg_cmp_dir (dir->i_sb->s_dev,
                                    LF (data)->u[i].child_off,
                                    d->d_name.name);
                if (rc == 0)
                        return LF (data)->u[i].child_off;
                else if (rc > 0)
                        return -rc;
        }
        return 0;
}

static long
ntreg_lookup_li (struct inode *dir, struct dentry *d, char *data)
{
        int i, rc;
        for (i = 0; i < LI (data)->num_child; i++) {
                rc = ntreg_cmp_dir (dir->i_sb->s_dev,
                                    LI (data)->u[i].child_off,
                                    d->d_name.name);
                if (rc == 0)
                        return LI (data)->u[i].child_off;
                else if (rc > 0)
                        return -rc;
        }
        return 0;
}

static long
ntreg_lookup_ri (struct inode *dir, struct dentry *d, char *data)
{
        int i, err;
        unsigned long ino = 0;
        char *sub_data = NULL;
        for (i = 0; i < RI (data)->num_child; i++) {
                sub_data = ntreg_read_node (dir->i_sb->s_dev, 
                                            RI (data)->u[i].child_off, &err);
                if (!sub_data)
                        return err;
                if (IS_LI (sub_data)) {
                        ino = ntreg_lookup_li (dir, d, sub_data);
                        if (ino)
                                break;
                } else {
                        ntreg_error ("Invalid RI entry for ino=0x%x",
                                     data);
                }
                ntreg_free (sub_data);
                sub_data = NULL;
        }
        if (sub_data)
                ntreg_free (sub_data);
        return ino;
}

static int
ntreg_lookup (struct inode *dir, struct dentry *d)
{
        int err;
        struct inode *res;
        long ino = 0;
	ntreg_debug(DEBUG_NAME1, "Looking up %s in %x\n",d->d_name.name,
                    (unsigned)dir->i_ino);

#if DO_SK
        if (strcmp (d->d_name.name, ".SD") == 0) {
                ino = NK_INO (dir)->sk_off;
        }
#endif
        if (!ino && (NK_INO (dir)->num_lf > 0))
        {
                int i;
                char *data = ntreg_read_node (dir->i_sb->s_dev, 
                                              NK_INO(dir)->lf_off, &err);
                if (!data)
                        return -err;
                if (IS_LF (data)) {
                        ino = ntreg_lookup_lf (dir, d, data);
                } else if (IS_LI (data)) {
                        ino = ntreg_lookup_li (dir, d, data);
                } else if (IS_RI (data)) {
                        ino = ntreg_lookup_ri (dir, d, data);
                } else {
                        ntreg_error ("Invalid leaf entry for ino=0x%x",
                                     dir->i_ino);
                }
                ntreg_free (data);
        }

        if (!ino && NK_INO (dir)->num_val > 0)
        {
                int i;
                long *val_arr = ntreg_read_node (dir->i_sb->s_dev, 
                                                 NK_INO(dir)->val_arr_off,
                                                 &err);
                if (!val_arr)
                        return -err;
                for (i = 0; i < NK_INO (dir)->num_val; i++) {
                        int rc;
                        rc = ntreg_cmp_file (dir->i_sb->s_dev,
                                             val_arr[i+1],
                                             d->d_name.name);
                        if (rc == 0) {
                                ino = val_arr[i+1];
                                break;
                        } else if (rc > 0) {
                                /* error */
                                ino = -rc;
                                break;
                        }
                }
                ntreg_free (val_arr);
        }

        if (ino < 0)
                return ino;

        res = 0;
        if (ino > 0) {
                res = iget (dir->i_sb, ino);
        }
        d_add (d, res);
        return 0;
}

static struct file_operations ntreg_file_operations = {
	NULL, /* lseek */
	ntreg_read,
	NULL, /* write */
	NULL, /* readdir */
	NULL, /* select */
	NULL, /* ioctl */
	NULL, /* mmap */
	NULL, /* open */
	NULL, /* flush */
	NULL, /* release */
	NULL, /* fsync */
	NULL, /* fasync */
	NULL, /* check_media_change */
	NULL, /* revalidate */
	NULL, /* lock */
};

static struct inode_operations ntreg_file_inode_operations = {
	&ntreg_file_operations,
	NULL, /* create */
	NULL, /* lookup */
	NULL, /* link */
	NULL, /* unlink */
	NULL, /* symlink */
	NULL, /* mkdir */
	NULL, /* rmdir */
	NULL, /* mknod */
	NULL, /* rename */
	NULL, /* readlink */
	NULL, /* follow_link */
	NULL, /* readpage */
	NULL, /* writepage */
	NULL, /* bmap */
	NULL, /* truncate */
	NULL, /* permission */
	NULL, /* smap */
	NULL, /* updatepage */
	NULL, /* revalidate */
};


static struct file_operations ntreg_dir_operations = {
	NULL, /* lseek */
	NULL, /* read */
	NULL, /* write */
	ntreg_readdir, /* readdir */
	NULL, /* poll */
	NULL, /* ioctl */
	NULL, /* mmap */
	NULL, /* open */
	NULL, /* flush */
	NULL, /* release */
	NULL, /* fsync */
	NULL, /* fasync */
	NULL, /* check_media_change */
	NULL, /* revalidate */
	NULL, /* lock */
};

static struct inode_operations ntreg_dir_inode_operations = {
	&ntreg_dir_operations,
	NULL,
	ntreg_lookup, /* lookup */
	NULL, /* link */
	NULL, /* unlink */
	NULL, /* symlink */
	NULL, /* mkdir */
	NULL, /* rmdir */
	NULL, /* mknod */
	NULL, /* rename */
	NULL, /* readlink */
	NULL, /* follow_link */
	NULL, /* readpage */
	NULL, /* writepage */
	NULL, /* bmap */
	NULL, /* truncate */
	NULL, /* permission */
	NULL, /* smap */
	NULL, /* updatepage */
	NULL, /* revalidate */
};

/* ntreg_read_inode is called by the Virtual File System (the kernel layer that
 * deals with filesystems) when iget is called requesting an inode not already
 * present in the inode table. Typically filesystems have separate
 * inode_operations for directories, files and symlinks.
 */
static void
ntreg_read_inode (struct inode* inode)
{
        int sz, err;
        char *data;
	inode->i_op=NULL;
	inode->i_mode=0;

	ntreg_debug (DEBUG_OTHER, "ntreg_read_inode %x\n",
                     (unsigned)inode->i_ino);

        if (!(data = ntreg_read_node (inode->i_sb->s_dev, inode->i_ino, &err))) {
                ntreg_error ("ntreg_read_inode: failed to read node %p\n", inode->i_ino);
                return;
        }
        inode->u.generic_ip = data;

	inode->i_uid=0;
	inode->i_gid=0;
	inode->i_nlink=1;

        sz = ((struct ntreg_basic *)data)->size;
        if (sz < 0)
                sz = 0 - sz;
        inode->i_size = sz;
	if (IS_NK (data)) {
		inode->i_op=&ntreg_dir_inode_operations;
		inode->i_mode=S_IFDIR|S_IRUGO|S_IXUGO;
	} else if (IS_VK (data)) {
		inode->i_op= &ntreg_file_inode_operations;
                inode->i_size = VK_INO (inode)->val_size;
                /* FIXME: this may be off by one:  make sure */
                if (inode->i_size < 0)
                        inode->i_size &= 0x7fffffff;

		inode->i_mode=S_IFREG|S_IRUGO;
                /*
                 * this is totally hokey, but it makes it easier to get
                 * the value's type.  Also, values can't have multiple
                 * links, so this field would be wasted, anyway.
                 */
                inode->i_nlink = VK_INO (inode)->val_type;
#if DO_SK
	} else if (IS_SK (data)) {
		inode->i_op= &ntreg_file_inode_operations;
		inode->i_mode= S_IFREG|S_IRUGO;
                inode->i_nlink = SK (data)->sk_links;
                inode->i_size = SK (data)->sk_size;
#endif
        } else {
                /* FIXME: ?? error ??  should be assert (FALSE); */
                ntreg_error ("ntreg_read_inode: unknown node type for ino=0x%x\n",
                             inode->i_ino);
        }

        return;
}

static void
ntreg_clear_inode (struct inode *ino)
{
	ntreg_debug (DEBUG_OTHER, "ntreg_clear_inode %lx\n", ino->i_ino);
        ntreg_free (ino->u.generic_ip);
	return;
}

/* Called when umounting a filesystem by do_umount() in fs/super.c */
static void
ntreg_put_super (struct super_block *sb)
{
	ntreg_debug (DEBUG_OTHER, "ntreg_put_super\n");
	MOD_DEC_USE_COUNT;
}

/* Called by the kernel when asking for stats */
static int
ntreg_statfs (struct super_block *sb, struct statfs *sf, int bufsize)
{
	ntreg_debug (DEBUG_OTHER, "ntreg_statfs\n");
        return -EINVAL;
}

/* Define the super block operation that are implemented */
static struct super_operations ntreg_super_operations = {
	ntreg_read_inode,
	NULL,
	NULL, /* put_inode */
	NULL, /* delete_inode */
	NULL, /* notify_change */
	ntreg_put_super,
	NULL, /* write_super */
	NULL, /* ntreg_statfs, */
        NULL, /* remount */
        ntreg_clear_inode,
};

/* Called to mount a filesystem by read_super() in fs/super.c
 * Return a super block, the main structure of a filesystem
 */
struct super_block *
ntreg_read_super(struct super_block *sb, void *options, int silent)
{
	struct buffer_head *bh;
	int i;

	MOD_INC_USE_COUNT;

	/* Don't put ntreg_debug() before MOD_INC_USE_COUNT, printk() can block
	 * so this could lead to a race condition with kmod.
	 */
	ntreg_debug (DEBUG_OTHER, "ntreg_read_super\n");

	/* Ensure that the super block won't be used until it is completed */
	lock_super (sb);
	ntreg_debug (DEBUG_OTHER, "lock_super\n");

	/* Set to read only */
	sb->s_flags |= MS_RDONLY;

	/* Assume a 512 bytes block device for now */
	set_blocksize (sb->s_dev, 512);
	/* Read the super block (boot block) */
	if(!(bh=bread (sb->s_dev, 0, 512))) {
		ntreg_error ("Reading super block failed\n");
		goto ntreg_read_super_fail;
	}
	ntreg_debug (DEBUG_OTHER, "Done reading boot block\n");

	/* Check for 'regf' signature */
	if (!NTREG_IS_REGF (bh->b_data)){
		ntreg_debug(DEBUG_OTHER, "Not a ntreg volume\n");
		brelse (bh);
		goto ntreg_read_super_fail;
	}
        brelse (bh); bh=NULL;
	sb->s_blocksize = 0x1000;
	for (i=sb->s_blocksize, sb->s_blocksize_bits=0; i != 1; i>>=1)
		sb->s_blocksize_bits++;
	set_blocksize (sb->s_dev, sb->s_blocksize);
	ntreg_debug (DEBUG_OTHER, "set_blocksize\n");

	/* Inform the kernel about which super operations are available */
	sb->s_op = &ntreg_super_operations;
	sb->s_magic = NTREG_SUPER_MAGIC;

	ntreg_debug (DEBUG_OTHER, "getting root dir\n");
	/* Get the root directory */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,13)
	if (!(sb->s_root=d_alloc_root (iget (sb, 0x20))))
#else
	if (!(sb->s_root=d_alloc_root (iget (sb, 0x20), NULL)))
#endif
        {
		ntreg_error ("Could not get root dir inode\n");
		goto ntreg_read_super_fail;
	}

	unlock_super (sb);
	ntreg_debug (DEBUG_OTHER, "unlock_super\n");
	ntreg_debug (DEBUG_OTHER, "read_super: done\n");
	return sb;

ntreg_read_super_fail:
	sb->s_dev = 0;
	unlock_super (sb);
	ntreg_debug (DEBUG_OTHER, "unlock_super\n");
	ntreg_debug (DEBUG_OTHER, "read_super: done\n");

	MOD_DEC_USE_COUNT;

	return NULL;
}

/*
 * Define the filesystem
 */
static struct file_system_type ntreg_fs_type = {
	"ntreg",
	FS_REQUIRES_DEV,
	ntreg_read_super,
	NULL
};

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,13)
int __init init_ntreg_fs (void)
#else
__initfunc (int init_ntreg_fs (void))
#endif
{
	printk (KERN_NOTICE "ntreg driver\n");
	ntreg_debug (DEBUG_OTHER, "registering %s\n",ntreg_fs_type.name);

        /* register the filesystem */
	return register_filesystem (&ntreg_fs_type);
}

#ifdef MODULE

/* No function of this module is needed by another module */
EXPORT_NO_SYMBOLS;


int
init_module (void)
{
	return init_ntreg_fs ();
}

void
cleanup_module (void)
{
	ntreg_debug (DEBUG_OTHER, "unregistering %s\n",ntreg_fs_type.name);
	unregister_filesystem (&ntreg_fs_type);
}
#endif

/*
 * Local variables:
 *  c-file-style: "linux"
 * End:
 */
